import glob
from .dataset import dataset

class Io:

    def read(path, batch_size=10):
        """
        Info:
            Lê um conjunto de imagens de um diretório.
        Params:
            path (string): Caminho do diretório.
            batch_size (int): Tamanho do lote de imagens.
        return:
            Objeto dataset.
        """
        batchs = {}
        files = glob.glob(path)

        if len(files) < batch_size:
            raise Exception("O tamanho do lote é maior que o número de imagens no diretório.")

        count = 0
        for batch in range(0, len(files), batch_size):
            batchs[str(count)] = files[batch:batch+batch_size]

            count += 1

        return dataset(batchs)

        

        





import numpy as np
from skimage.color import rgb2gray

class Preprocessing:

    @staticmethod
    def normalize(dataset):
        """
        Info:
            Normaliza os valores de um dataset.
        Params:
            dataset (array numpy): Dataset com imagens 2D-dimensional.
        Return:
            Dataset normalizado.
        """
        normalized = []

        for img in dataset:
            normalized.append((img - np.min(img)) / (np.max(img) - np.min(img)))

        return np.asarray(normalized)

    @staticmethod
    def label_encoder(names, class_1, class_2):
        """
        Info:
            Codifica os valores de um dataset entre 0 e 1.
        Params:
            dataset (array numpy): Dataset com o nome das classes de cada imagem.
        Return:
            Dataset codificado.
        """
        encoded = []

        for name in names:
        
            if name.lower() == class_1.lower():
                encoded.append([1])

            elif name.lower() == class_2.lower():
                encoded.append([0])

        return np.asarray(encoded)

    @staticmethod
    def to_gray(dataset):
        """
        Info:
            Converte um dataset de imagens RGB para escala de cinza.
        Params:
            dataset (array numpy): Dataset com imagens 3D-dimensional.
        Return:
            Dataset convertido.
        """
        gray = []

        for img in dataset:
            gray.append((rgb2gray(img)))

        return np.asarray(gray)

    @staticmethod
    def convert_to_integer(dataset):
        """
        Info:
            Converte um dataset de imagens float para int entre 0 e 255.
        Params:
            dataset (array numpy): Dataset com imagens 2D-dimensional.
        Return:
            Dataset convertido.
        """
        integer = []

        for img in dataset:
            integer.append((img*255).astype(np.uint8))

        return np.asarray(integer)

    @staticmethod
    def contrast_flare(dataset, k, e):
        """
        Info:
            Aplica o alargamento de contraste nas imagens de um dataset.
        Params:
            dataset (array numpy): Dataset com imagens 2D-dimensional no intervalo de 0 a 255.
            k (float): Valor de contraste.
            e (float): Valor de exposição.
        """
        contrast = []
        for img in dataset:
            contrast.append(1/(1+(k/img)**e))
        
        return np.asarray(contrast)

    @staticmethod
    def negatives(dataset):
        """
        Info:
            Aplica o negativo nas imagens de um dataset.
        Params:
            dataset (array numpy): Dataset com imagens 2D-dimensional no intervalo de 0 a 255.
        """
        negatives = []
        for img in dataset:
            negatives.append(255-img)
        
        return np.asarray(negatives)

    def log_transform(dataset, c):
        """
        Info:
            Aplica a transformação logarítmica nas imagens de um dataset.
        Params:
            dataset (array numpy): Dataset com imagens 2D-dimensional no intervalo de 0 a 255.
            c (float): Valor de constante.
        """
        log = []
        for img in dataset:
            log.append(c*np.log(1+img))
        
        return np.asarray(log)

    def gamma_transform(dataset, c, g):
        """
        Info:
            Aplica a transformação gamma nas imagens de um dataset.
        Params:
            dataset (array numpy): Dataset com imagens 2D-dimensional no intervalo de 0 a 255.
            c (float): Valor de constante.
            g (float): Valor de gamma.
        """
        gamma = []
        for img in dataset:
            gamma.append(c*(img**g))
        
        return np.asarray(gamma)

    


import numpy as np

class Filters:

    @staticmethod
    def conv2d(img, kernel):
        """
        Info:
            Aplica um filtro convolucional em uma imagem.
        Params:
            img (array numpy): Imagem 2D-dimensional no intervalo de 0 a 255.
            kernel (array numpy): Kernel 2D-dimensional.
        """
        # Aplica o filtro convolucional
        #Aplicando a convolucao na imagem.
        Filters._check_dimensions(img)
        Filters._check_dimensions_kernel(kernel)
        
        new_pixels = np.zeros(img.shape)
        bord = int(np.floor(kernel.shape[0]/2))

        for lines in range(img.shape[0] - (kernel.shape[0]-1)):
            for columns in range(img.shape[1] - (kernel.shape[1]-1)):
                new_pixels[lines + bord][columns + bord] = np.sum(kernel*img[lines:lines+kernel.shape[0], columns:columns+kernel.shape[1]])

        return new_pixels

    @staticmethod
    def vertical_sobel(img):
        """
        Info:
            Aplica o filtro de Sobel em uma imagem.
        Params:
            img (array numpy): Imagem 2D-dimensional no intervalo de 0 a 255.
        """
        # Aplica o filtro de Sobel
        kernel = np.array([[-1, 0, 1], 
                           [-2, 0, 2], 
                           [-1, 0, 1]])
        sobel = Filters.conv2d(img, kernel)
        return sobel

    @staticmethod
    def horizontal_sobel(img):
        """
        Info:
            Aplica o filtro de Sobel em uma imagem.
        Params:
            img (array numpy): Imagem 2D-dimensional no intervalo de 0 a 255.
        """
        # Aplica o filtro de Sobel
        kernel = np.array([[-1, -2, -1], 
                           [ 0,  0,  0], 
                           [ 1,  2,  1]])
        sobel = Filters.conv2d(img, kernel)
        return sobel

    @staticmethod
    def laplace(img):
        """
        Info:
            Aplica o filtro de Laplace em uma imagem.
        Params:
            img (array numpy): Imagem 2D-dimensional no intervalo de 0 a 255.
        """
        # Aplica o filtro de Laplace
        kernel = np.array([[0,  1, 0], 
                           [1, -4, 1], 
                           [0,  1, 0]])
        laplace = Filters.conv2d(img, kernel)
        return laplace

    @staticmethod
    def horizontal_prewitt(img):
        """
        Info:
            Aplica o filtro de Prewitt em uma imagem.
        Params:
            img (array numpy): Imagem 2D-dimensional no intervalo de 0 a 255.
        """
        # Aplica o filtro de Prewitt
        kernel = np.array([[ 1/3,   1/3,   1/3], 
                           [   0,     0,     0], 
                           [-1/3,  -1/3,  -1/3]])
        prewitt = Filters.conv2d(img, kernel)
        return prewitt

    @staticmethod
    def vertical_prewitt(img):
        """
        Info:
            Aplica o filtro de Prewitt em uma imagem.
        Params:
            img (array numpy): Imagem 2D-dimensional no intervalo de 0 a 255.
        """
        # Aplica o filtro de Prewitt
        kernel = np.array([[ 1/3,     0,  -1/3], 
                           [ 1/3,     0,  -1/3], 
                           [ 1/3,     0,  -1/3]])
        prewitt = Filters.conv2d(img, kernel)
        return prewitt

    @staticmethod
    def horizontal_scharr(img):
        """
        Info:
            Aplica o filtro de Scharr em uma imagem.
        Params:
            img (array numpy): Imagem 2D-dimensional no intervalo de 0 a 255.
        """
        # Aplica o filtro de Scharr
        kernel = np.array([[ 3,  10,  3], 
                           [ 0,   0,  0], 
                           [-3, -10, -3]])
        scharr = Filters.conv2d(img, kernel)
        return scharr
    
    @staticmethod
    def vertical_scharr(img):
        """
        Info:
            Aplica o filtro de Scharr em uma imagem.
        Params:
            img (array numpy): Imagem 2D-dimensional no intervalo de 0 a 255.
        """
        # Aplica o filtro de Scharr
        kernel = np.array([[ 3,  0,  -3], 
                           [10,  0, -10], 
                           [ 3,  0,  -3]])
        scharr = Filters.conv2d(img, kernel)
        return scharr

    @staticmethod
    def median(img, kernel_size):
        """
        Info:
            Aplica o filtro de mediana em uma imagem.
        Params:
            img (array numpy): Imagem 2D-dimensional no intervalo de 0 a 255.
            kernel_size (int): Tamanho do kernel.
        """
        # Aplica o filtro de mediana
        median = Filters._convNoLinear(img, kernel_size, 'median')
        return median 

    @staticmethod
    def maximum(img, kernel_size):
        """
        Info:
            Aplica o filtro de máximo em uma imagem.
        Params:
            img (array numpy): Imagem 2D-dimensional no intervalo de 0 a 255.
            kernel_size (int): Tamanho do kernel.
        """
        # Aplica o filtro de máximo
        maximum = Filters._convNoLinear(img, kernel_size, 'max')
        return maximum

    @staticmethod
    def minimum(img, kernel_size):
        """
        Info:
            Aplica o filtro de mínimo em uma imagem.
        Params:
            img (array numpy): Imagem 2D-dimensional no intervalo de 0 a 255.
            kernel_size (int): Tamanho do kernel.
        """
        # Aplica o filtro de mínimo
        minimum = Filters._convNoLinear(img, kernel_size, 'min')
        return minimum
    
    @staticmethod
    def mean(img, kernel_size):
        """
        Info:
            Aplica o filtro de média em uma imagem.
        Params:
            img (array numpy): Imagem 2D-dimensional no intervalo de 0 a 255.
            kernel_size (int): Tamanho do kernel.
        """
        # Aplica o filtro de média
        mean = Filters._convNoLinear(img, kernel_size, 'mean')
        return mean

    def _convNoLinear(img, kernel_shape, operation):
        """
        Info:
            Aplica um filtro convolucional não linear em uma imagem.
        Params:
            img (array numpy): Imagem 2D-dimensional no intervalo de 0 a 255.
            kernel_shape (int): Dimensão do kernel.
            operation (str): Operação a ser realizada no kernel.
        """
        operations = {
            'max': np.max,
            'min': np.min,
            'mean': np.mean,
            'median': np.median
        }

        if kernel_shape % 2 == 0:
            raise ValueError('O tamanho do kernel deve ser ímpar.')

        Filters._check_dimensions(img)
        
        new_pixels = np.zeros(img.shape)
        bord = int(np.floor(kernel_shape/2))

        for lines in range(img.shape[0] - (kernel_shape-1)):
            for columns in range(img.shape[1] - (kernel_shape-1)):
                new_pixels[lines + bord][columns + bord] = operations[operation](img[lines:lines+kernel_shape, columns:columns+kernel_shape])

        return new_pixels
    
    def _check_dimensions(img):
        if len(img.shape) == 2:
            return True
        else:
            raise Exception("A imagem deve ser 2D-dimensional.")

    def _check_dimensions_kernel(kernel):
        if len(kernel.shape) == 2 and kernel.shape[0] == kernel.shape[1] and kernel.shape[0] % 2 != 0:
            return True
        else:
            raise Exception("O kernel deve ser 2D-dimensional, quadricular e com dimensões impar.")

    



from skimage.io import imread
import numpy as np

class dataset:

    def __init__(self, batches):
        self._files = []
        self._batches = batches
        self._n_batches = len(batches)
        self._read_files()

    def read_batch(self, batch):
        """
        Info:
            Lê um lote de imagens do dataset.
        Params:
            batch (int): Id do lote.
        Return:
            Lista de imagens.
        """
        images = []
        for file in self._batches[str(batch)]:
            images.append(imread(file))
        
        return np.asarray(images)


    def _read_files(self):
        """
        Info:
            Lê todos os diretórios dos arquivos do dataset.
        Return:
            Lista com diretórios.
        """
        for batch in self._batches:
            for file in self._batches[batch]:
                self._files.append(file)

        self._files = np.asarray(self._files)

    @property
    def files(self):
        return self._files

    @property
    def n_batches(self):
        return self._n_batches






import random
import numpy as np
from .model_tools import Tools

class MClassifier:
    """
    Info:
        Rede neural baseada no modelo perceptron de multicamadas (Multi-Layer Perceptron - MLP).
        Por enquanto, a rede neural possui apenas uma camada de entrada, uma oculta e uma de saida.

    Functions: 
        fit: Função para treinar a rede neural.
        predict: Função para predizer o resultado da rede neural.
        initialize_weights: Função para inicializar os pesos da rede neural.
    
    """

    def __init__(self):
        #Inicializa listas para armazenar os pesos.
        self._layers = {}
        self._layers_activated = {}
        self._n_layers = 2
        #Inicializa uma variável para armazenar o erro.
        self._mean_absolute_loss = 0
        #Metrics
        self._accuracy = 0
        self._precision = 0
        self._recall = 0
        self._f1 = 0
        self._kappa = 0

    def fit(self, x, y, epoch=5, neurons=5, n_layers=2, learning_rate=0.0001, error_threshold=0.1, moment=1):
        """
        Info: 
            Esta função realiza o treinamento da rede neural.
            A rede será treinada até que o valor máximo para o loss seja atingido,
            até que seja atingido o número de épocas definido.
        Params:
            x (Array numpy): Valores de entrada / features.
            y (array numpy): Rotulos corretamente marcados.
            epoch (int): Valor de epocas de treinamento.
            neurons (int): Representa a quantidade de neuronios.
            n_layers (int): Representa a quantidade de camadas. 
            learning_rate (float): Taxa de aprendizagem da rede.
            error_threshold (float): Valor máximo para o loss.
            moment (float): Otimiza o aprendizado evitantando mínimos locais na curva de erro.
        """
        self.check_params(epoch, neurons, n_layers, learning_rate)

        #--------------------------INICIALIZANDO OS PESOS------------------------------
        self._n_layers = n_layers
        self.initialize_weights(x, neurons, n_layers)
        PREFIX = "Layer-"
        #--------------------------INICIALIZANDO O TREINO------------------------------
        #Percorre o numero de epocas
        for epc in range(epoch):
            #Atribui as features para a camada de entrada
            layer_active = x

            for id_layer in range(n_layers):
                layer = np.dot(layer_active, self._layers[PREFIX + str(id_layer)])
                layer_active = self.sigmoid(layer)
                self._layers_activated[PREFIX + str(id_layer)] = layer_active

            #------------------------CALCULANDO O ERRO---------------------------------
            loss = y - layer_active
            self._mean_absolute_loss = np.mean(np.abs(loss))
            print("Loss: {}".format(self._mean_absolute_loss))

            if self._mean_absolute_loss <= error_threshold:
                y_pred = [[self.thresh(i)] for i in layer_active]
                self._accuracy, self._precision, self._recall, self._f1, self._kappa = Tools.metrics(y, y_pred)
                break
            
            #---------------------ATUALIZANDO OS PESOS-------------------------------
            #Último layer
            derivative_last_layer = self.sigmoid_derivative(self._layers_activated[PREFIX + str(n_layers - 1)])
            delta_last_layer = loss * derivative_last_layer

            #Atualizando os pesos da ultima camada
            activeXdelta = np.dot(self._layers_activated[PREFIX + str(n_layers - 2)].T, delta_last_layer)
            self._layers[PREFIX + str(n_layers - 1)] = (self._layers[PREFIX + str(n_layers - 1)] * moment) + (activeXdelta * learning_rate)

            #Layers ocultos
            for id_layer in range(n_layers - 2, 0, -1):
                delta_last_layer = np.dot(delta_last_layer, self._layers[PREFIX + str(id_layer + 1)].T) * self.sigmoid_derivative(self._layers_activated[PREFIX + str(id_layer)])
                self._layers[PREFIX + str(id_layer)] = (self._layers[PREFIX + str(id_layer)] * moment) + (np.dot(self._layers_activated[PREFIX + str(id_layer - 1)].T, delta_last_layer) * learning_rate)

            #Camada de entrada
            delta_last_layer = np.dot(delta_last_layer, self._layers[PREFIX + str(1)].T) * self.sigmoid_derivative(self._layers_activated[PREFIX + str(0)])
            self._layers[PREFIX + str(0)] = (self._layers[PREFIX + str(0)] * moment) + (np.dot(x.T, delta_last_layer) * learning_rate)

        #-------------------------FIM DO TREINO-------------------------------------------

        #---------------------CALCULANDO MÉTRICAS-----------------------------------------
        y_pred = [[self.thresh(i)] for i in layer_active]
        self._accuracy, self._precision, self._recall, self._f1, self._kappa = Tools.metrics(y, y_pred)
             
    def predict(self, x):
        """
        Info:
            Camada de predição da rede neural.
        Params:
            x (Array numpy): Valores de entrada / features.
        """

        #Inicia uma lista para armazenar as predições
        predict = []
        PREFIX = "Layer-"
        #--------------------------INICIALIZANDO O PREDIÇÃO------------------------------
        #Percorre feature por feature
        for i in range(len(x)):

            layer_active = x[i]
            
            for id_layer in range(self._n_layers):
                layer = np.dot(layer_active, self._layers[PREFIX + str(id_layer)])
                layer_active = self.sigmoid(layer)

            #Predição
            predict.append([self.thresh(layer_active)])

        return predict

    def initialize_weights(self, x, neurons, n_layers):
        """
        Info:
            Esta função inicializa os pesos das camadas da rede neural.

        Params:
            x: Features de entrada da rede.
            neurons: Quantidade de neurônios definidos para a camada oculta.
        """

        PREFIX = "Layer-"

        #Percorre o numero de caracteristicas
        generated_weights = []
        #Camada de entrada
        for _ in range(len(x[0])):

            #Gera uma lista de valores aleatorios
            generated_weights.append([round(random.uniform(-1, 1), 3) for _ in range(neurons)])
        
        self._layers[PREFIX+'0'] = np.asarray(generated_weights)

        #Camada oculta
        for layer in range(1, n_layers-1):
            self._weights_layer2 = []

            for _ in range(neurons):
                generated_weights = [round(random.uniform(-1, 1), 3) for _ in range(neurons)]
                self._weights_layer2.append(generated_weights)

            self._layers[PREFIX + str(layer)] = np.asarray(self._weights_layer2)

        #Camada de saida
        self._weights_layer2 = []
        for _ in range(neurons):
            generated_weights = [round(random.uniform(-1, 1), 3)]
            self._weights_layer2.append(generated_weights)

        self._layers[PREFIX + str(n_layers-1)] = np.asarray(self._weights_layer2)
    
    def sigmoid(self, value):
        """
        Info:
            Função para aplicar a função sigmoid.
        Params:
            value (float): Valor para aplicar a função.
        """
        return 1 / (1 + np.exp(-value))
    
    def sigmoid_derivative(self, value):
        """
        Info:
            Função para aplicar a derivada da função sigmoid.
        Params:
            value (float): Valor para aplicar a derivada.
        """
        return value * (1 - value)

    def thresh(self, value):
        """
        Info:
            Função para aplicar a função de threshold.
        Params:
            value (float): Valor para aplicar a função.
        """
        return 1 if value >= 0.5 else 0

    def check_params(self, epoch, neurons, n_layers, learning_rate):
        """
        Info:
            Função para verificar os parâmetros de entrada.
        Params:
            epoch (int): Número de épocas.
            neurons (int): Número de neurônios.
            n_layers (int): Número de camadas.
            learning_rate (float): Taxa de aprendizado.
        """
        if epoch < 1:
            raise ValueError("O número de épocas deve ser maior que 0.")
        if neurons < 1:
            raise ValueError("O número de neurônios deve ser maior que 0.")
        if n_layers < 2:
            raise ValueError("O número de camadas deve ser maior que 1.")
        if learning_rate <= 0:
            raise ValueError("A taxa de aprendizado deve ser maior que 0.")



import pickle
from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score, cohen_kappa_score

class Tools:

    """Esta classe tem como objetivo facilitar a criação de modelos e a avaliação deles."""

    @staticmethod
    def metrics(y_true, y_pred):
        """
        Info:
            Calcula as métricas de acurácia, precisão, recall, f1 e kappa.
        Params:
            y_true (array numpy): Valores corretamente marcados.
            y_pred (array numpy): Valores preditos.
        """
        accuracy = accuracy_score(y_true, y_pred)
        precision = precision_score(y_true, y_pred)
        recall = recall_score(y_true, y_pred)
        f1 = f1_score(y_true, y_pred)
        kappa = cohen_kappa_score(y_true, y_pred)

        return accuracy, precision, recall, f1, kappa

    @staticmethod
    def save(model, filename):
        """
        Info:
            Salva o modelo em um arquivo pkl.
        Params:
            model (object model): Modelo a ser salvo.
            filename (string): Path com o destino e nome do modelo, adicionar a extensão .pkl
        """
        with open(filename, 'wb') as f:
            pickle.dump(model, f)

    @staticmethod
    def load(filename):
        """
        Info:
            Carrega o modelo salvo.
        Params:
            filename (string): Path com o caminho para o modelo.
        """
        with open(filename, 'rb') as f:
            return pickle.load(f)



from skimage.feature import graycomatrix, graycoprops
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
import numpy as np

class Features:

    @staticmethod
    def GLCM(dataset, steps = 1):
        """
        Info:
            Calcula a matriz de Co-ocorrencia de níveis de cinza de um dataset.
        Params:
            dataset (array numpy, array, imread_collection): Dataset de imagens.
            steps (array): Lista de deslocamentos de distância de pares de pixels.
        Return:
            Matriz isotrópica.
        """
        matrix = []

        for img in dataset:
            matrix0 = graycomatrix(img, [steps], [0], normed=True)
            matrix1 = graycomatrix(img, [steps], [np.pi/4], normed=True)
            matrix2 = graycomatrix(img, [steps], [np.pi/2], normed=True)
            matrix3 = graycomatrix(img, [steps], [3*np.pi/4], normed=True)
            matrix.append((matrix0+matrix1+matrix2+matrix3)/4)
        
        return np.asarray(matrix)
    
    @staticmethod
    def GLCM_props(matrix):
        """
        Info:
            Calcula as propriedades da matriz de Co-ocorrencia de níveis de cinza.
        Params:
            matrix (array): Array de matrizes de Co-ocorrencia de níveis de cinza.
        Return:
            Matriz com as propriedades de contraste, dissimilaridade, homogeneidade, 
            energia, correlação e ASM.
        """
        props = []

        for mat in matrix:
            prop = np.zeros((6))
            prop[0] = graycoprops(mat,'contrast')
            prop[1] = graycoprops(mat,'dissimilarity')
            prop[2] = graycoprops(mat,'homogeneity')
            prop[3] = graycoprops(mat,'energy')
            prop[4] = graycoprops(mat,'correlation')
            prop[5] = graycoprops(mat,'ASM')
            props.append(prop)
        
        return np.asarray(props)

    @staticmethod
    def categorize(dataset, new_value_bit, old_value_bit):
        """
        Info:
            Categoriza os bits de imagens de um dataset.
        Params:
            dataset (array numpy): Dataset de imagens bi-dimensionais.
            new_value_bit (int): Novo valor de bits.
            old_value_bit (int): Valor dos bits da imagem
        Return:
            Array com as imagens categorizadas.
        """
        categorized = []

        for img in dataset:

            #Guardando as dimensões da imagem original.
            Lines,Columns = img.shape;

            img = np.array(img, dtype=np.float64)

            #Categorizando os bits.
            for line in range(Lines):
                for Colum in range(Columns):
                    img[line, Colum] = (((2**new_value_bit)-1) * img[line, Colum]) // ((2**old_value_bit)-1)

            img = np.array(img, dtype=np.uint8)

            categorized.append(img)

        return np.asarray(categorized)
    @staticmethod
    def PCA(X, n_components):
        """
        Info:
            Aplica o algoritmo de redução de dimensionalidade PCA em uma matriz de características.
        Params:
            X (array): Matriz de características.
            n_components (int): Número de componentes principais.
        Return:
            Matriz de características reduzida.
        """
        
        X = StandardScaler().fit_transform(X)

        pca = PCA(n_components=n_components)
        components = pca.fit_transform(X)
        
        return components


from setuptools import setup

with open("README.md", "r") as fh:
    readme = fh.read()

setup(name='keros',
    version='0.1.3',
    license='MIT License',
    author='Armando Luz Borges',
    long_description=readme,
    long_description_content_type="text/markdown",
    author_email='armandoari.288@gmail.com',
    description=u'Pacote destinado ao uso de conceitos de machine learning',
    packages=['keros'],
    install_requires=['numpy', 'scikit-learn', 'scikit-image', 'rich'])
